/*  Lziprecover - Data recovery tool for lzipped files
    Copyright (C) 2009, 2010, 2011, 2012 Antonio Diaz Diaz.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
    Return values: 0 for a normal exit, 1 for environmental problems
    (file not found, invalid flags, I/O errors, etc), 2 to indicate a
    corrupt or invalid input file, 3 for an internal consistency error
    (eg, bug) which caused lziprecover to panic.
*/

#define _FILE_OFFSET_BITS 64

#include <algorithm>
#include <cerrno>
#include <climits>
#include <csignal>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <string>
#include <vector>
#include <fcntl.h>
#include <stdint.h>
#include <unistd.h>
#include <utime.h>
#include <sys/stat.h>
#if defined(__MSVCRT__)
#define fchmod(x,y) 0
#define fchown(x,y,z) 0
#define SIGHUP SIGTERM
#define S_ISSOCK(x) 0
#define S_IRGRP 0
#define S_IWGRP 0
#define S_IROTH 0
#define S_IWOTH 0
#endif

#include "arg_parser.h"
#include "lzip.h"
#include "decoder.h"

#if CHAR_BIT != 8
#error "Environments where CHAR_BIT != 8 are not supported."
#endif


namespace {

const char * const Program_name = "Lziprecover";
const char * const program_name = "lziprecover";
const char * const program_year = "2012";
const char * invocation_name = 0;

#ifdef O_BINARY
const int o_binary = O_BINARY;
#else
const int o_binary = 0;
#endif

struct { const char * from; const char * to; } const known_extensions[] = {
  { ".lz",  ""     },
  { ".tlz", ".tar" },
  { 0,      0      } };

enum Mode { m_none, m_decompress, m_generate, m_list, m_merge, m_range,
            m_recover, m_repair, m_split, m_test, m_update };

std::string output_filename;
int outfd = -1;
const mode_t usr_rw = S_IRUSR | S_IWUSR;
const mode_t all_rw = usr_rw | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
mode_t outfd_mode = usr_rw;
bool delete_output_on_interrupt = false;


void show_help()
  {
  std::printf( "%s - Data recovery tool and decompressor for lzipped files.\n", Program_name );
  std::printf( "\nUsage: %s [options] [files]\n", invocation_name );
  std::printf( "\nOptions:\n"
               "  -h, --help                      display this help and exit\n"
               "  -V, --version                   output version information and exit\n"
               "  -c, --stdout                    send decompressed output to standard output\n"
               "  -d, --decompress                decompress\n"
               "  -D, --range-decompress=<range>  decompress only a range of bytes (N-M)\n"
               "  -f, --force                     overwrite existing output files\n"
//               "  -g, --generate-recover-file     generate a recover file\n"
               "  -k, --keep                      keep (don't delete) input files\n"
               "  -l, --list                      print total file sizes and ratios\n"
               "  -m, --merge                     correct errors in file using several copies\n"
               "  -o, --output=<file>             place the output into <file>\n"
               "  -q, --quiet                     suppress all messages\n"
//               "  -r, --recover                   correct errors in file using a recover file\n"
               "  -R, --repair                    try to repair a small error in file\n"
               "  -s, --split                     split multi-member file in single-member files\n"
               "  -t, --test                      test compressed file integrity\n"
//               "  -u, --update                    convert file from version 0 to version 1\n"
               "  -v, --verbose                   be verbose (a 2nd -v gives more)\n"
               "Numbers may be followed by a multiplier: k = kB = 10^3 = 1000,\n"
               "Ki = KiB = 2^10 = 1024, M = 10^6, Mi = 2^20, G = 10^9, Gi = 2^30, etc...\n"
               "\nReport bugs to lzip-bug@nongnu.org\n"
               "Lziprecover home page: http://www.nongnu.org/lzip/lziprecover.html\n" );
  }


void show_version()
  {
  std::printf( "%s %s\n", Program_name, PROGVERSION );
  std::printf( "Copyright (C) %s Antonio Diaz Diaz.\n", program_year );
  std::printf( "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n"
               "This is free software: you are free to change and redistribute it.\n"
               "There is NO WARRANTY, to the extent permitted by law.\n" );
  }


void one_file( const int argind, const int arguments )
  {
  if( argind + 1 != arguments )
    {
    show_error( "You must specify exactly 1 file.", 0, true );
    std::exit( 1 );
    }
  }


void set_mode( Mode & program_mode, const Mode new_mode )
  {
  if( program_mode != m_none && program_mode != new_mode )
    {
    show_error( "Only one operation can be specified.", 0, true );
    std::exit( 1 );
    }
  program_mode = new_mode;
  }


int extension_index( const std::string & name )
  {
  for( int i = 0; known_extensions[i].from; ++i )
    {
    const std::string ext( known_extensions[i].from );
    if( name.size() > ext.size() &&
        name.compare( name.size() - ext.size(), ext.size(), ext ) == 0 )
      return i;
    }
  return -1;
  }


void set_d_outname( const std::string & name, const int i )
  {
  if( i >= 0 )
    {
    const std::string from( known_extensions[i].from );
    if( name.size() > from.size() )
      {
      output_filename.assign( name, 0, name.size() - from.size() );
      output_filename += known_extensions[i].to;
      return;
      }
    }
  output_filename = name; output_filename += ".out";
  if( verbosity >= 1 )
    std::fprintf( stderr, "%s: Can't guess original name for '%s' -- using '%s'.\n",
                  program_name, name.c_str(), output_filename.c_str() );
  }


bool open_outstream( const bool force )
  {
  int flags = O_CREAT | O_WRONLY | o_binary;
  if( force ) flags |= O_TRUNC; else flags |= O_EXCL;

  outfd = open( output_filename.c_str(), flags, outfd_mode );
  if( outfd < 0 && verbosity >= 0 )
    {
    if( errno == EEXIST )
      std::fprintf( stderr, "%s: Output file '%s' already exists, skipping.\n",
                    program_name, output_filename.c_str() );
    else
      std::fprintf( stderr, "%s: Can't create output file '%s': %s.\n",
                    program_name, output_filename.c_str(), std::strerror( errno ) );
    }
  return ( outfd >= 0 );
  }


bool check_tty( const int infd )
  {
  if( isatty( infd ) )
    {
    show_error( "I won't read compressed data from a terminal.", 0, true );
    return false;
    }
  return true;
  }


void cleanup_and_fail( const int retval )
  {
  if( delete_output_on_interrupt )
    {
    delete_output_on_interrupt = false;
    if( verbosity >= 0 )
      std::fprintf( stderr, "%s: Deleting output file '%s', if it exists.\n",
                    program_name, output_filename.c_str() );
    if( outfd >= 0 ) { close( outfd ); outfd = -1; }
    if( std::remove( output_filename.c_str() ) != 0 && errno != ENOENT )
      show_error( "WARNING: deletion of output file (apparently) failed." );
    }
  std::exit( retval );
  }


     // Set permissions, owner and times.
void close_and_set_permissions( const struct stat * const in_statsp )
  {
  bool warning = false;
  if( in_statsp )
    {
    // fchown will in many cases return with EPERM, which can be safely ignored.
    if( ( fchown( outfd, in_statsp->st_uid, in_statsp->st_gid ) != 0 &&
          errno != EPERM ) ||
        fchmod( outfd, in_statsp->st_mode ) != 0 ) warning = true;
    }
  if( close( outfd ) != 0 ) cleanup_and_fail( 1 );
  outfd = -1;
  delete_output_on_interrupt = false;
  if( in_statsp )
    {
    struct utimbuf t;
    t.actime = in_statsp->st_atime;
    t.modtime = in_statsp->st_mtime;
    if( utime( output_filename.c_str(), &t ) != 0 ) warning = true;
    }
  if( warning && verbosity >= 1 )
    show_error( "Can't change output file attributes." );
  }


std::string insert_fixed( std::string name )
  {
  if( name.size() > 4 && name.compare( name.size() - 4, 4, ".tlz" ) == 0 )
    name.insert( name.size() - 4, "_fixed" );
  else if( name.size() > 3 && name.compare( name.size() - 3, 3, ".lz" ) == 0 )
    name.insert( name.size() - 3, "_fixed" );
  else name += "_fixed.lz";
  return name;
  }


unsigned char xdigit( const int value )
  {
  if( value >= 0 && value <= 9 ) return '0' + value;
  if( value >= 10 && value <= 15 ) return 'A' + value - 10;
  return 0;
  }


void show_trailing_garbage( const uint8_t * const data, const int size,
                            const Pretty_print & pp, const bool all )
  {
  std::string garbage_msg;
  if( !all ) garbage_msg = "first bytes of ";
  garbage_msg += "trailing garbage found = ";
  bool text = true;
  for( int i = 0; i < size; ++i )
    if( !std::isprint( data[i] ) ) { text = false; break; }
  if( text )
    {
    garbage_msg += '\'';
    garbage_msg.append( (const char *)data, size );
    garbage_msg += '\'';
    }
  else
    {
    for( int i = 0; i < size; ++i )
      {
      if( i > 0 ) garbage_msg += ' ';
      garbage_msg += xdigit( data[i] >> 4 );
      garbage_msg += xdigit( data[i] & 0x0F );
      }
    }
  pp( garbage_msg.c_str() );
  }


int decompress( const int infd, const Pretty_print & pp, const bool testing )
  {
  int retval = 0;

  try {
    Range_decoder rdec( infd );
    long long partial_file_pos = 0;
    for( bool first_member = true; ; first_member = false, pp.reset() )
      {
      File_header header;
      rdec.reset_member_position();
      const int size = rdec.read( header.data, File_header::size );
      if( rdec.finished() )			// End Of File
        {
        if( first_member )
          { pp( "Error reading member header" ); retval = 1; }
        else if( verbosity >= 4 && size > 0 )
          show_trailing_garbage( header.data, size, pp, true );
        break;
        }
      if( !header.verify_magic() )
        {
        if( first_member )
          { pp( "Bad magic number (file not in lzip format)" ); retval = 2; }
        else if( verbosity >= 4 )
          show_trailing_garbage( header.data, size, pp, false );
        break;
        }
      if( !header.verify_version() )
        {
        if( verbosity >= 0 )
          { pp();
            std::fprintf( stderr, "Version %d member format not supported.\n",
                          header.version() ); }
        retval = 2; break;
        }
      if( header.dictionary_size() < min_dictionary_size ||
          header.dictionary_size() > max_dictionary_size )
        { pp( "Invalid dictionary size in member header" ); retval = 2; break; }

      if( verbosity >= 2 || ( verbosity == 1 && first_member ) )
        {
        pp();
        if( verbosity >= 2 )
          std::fprintf( stderr, "version %d, dictionary size %7sB.  ",
                        header.version(),
                        format_num( header.dictionary_size(), 9999, -1 ) );
        }

      LZ_decoder decoder( header, rdec, outfd );
      const int result = decoder.decode_member( pp );
      partial_file_pos += rdec.member_position();
      if( result != 0 )
        {
        if( verbosity >= 0 && result <= 2 )
          {
          pp();
          if( result == 2 )
            std::fprintf( stderr, "File ends unexpectedly at pos %lld\n",
                          partial_file_pos );
          else
            std::fprintf( stderr, "Decoder error at pos %lld\n",
                          partial_file_pos );
          }
        retval = 2; break;
        }
      if( verbosity >= 2 )
        { if( testing ) std::fprintf( stderr, "ok\n" );
          else std::fprintf( stderr, "done\n" ); }
      }
    }
  catch( std::bad_alloc )
    {
    pp( "Not enough memory. Find a machine with more memory" );
    retval = 1;
    }
  catch( Error e ) { pp(); show_error( e.msg, errno ); retval = 1; }
  if( verbosity == 1 && retval == 0 )
    { if( testing ) std::fprintf( stderr, "ok\n" );
      else std::fprintf( stderr, "done\n" ); }
  return retval;
  }


extern "C" void signal_handler( int )
  {
  show_error( "Control-C or similar caught, quitting." );
  cleanup_and_fail( 1 );
  }


void set_signals()
  {
  std::signal( SIGHUP, signal_handler );
  std::signal( SIGINT, signal_handler );
  std::signal( SIGTERM, signal_handler );
  }

} // end namespace


int verbosity = 0;


const char * format_num( long long num, long long limit,
                         const int set_prefix )
  {
  const char * const si_prefix[8] =
    { "k", "M", "G", "T", "P", "E", "Z", "Y" };
  const char * const binary_prefix[8] =
    { "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi" };
  static bool si = true;
  static char buf[32];

  if( set_prefix ) si = ( set_prefix > 0 );
  const int factor = ( si ? 1000 : 1024 );
  const char * const * prefix = ( si ? si_prefix : binary_prefix );
  const char * p = "";
  bool exact = ( num % factor == 0 );

  for( int i = 0; i < 8 && ( llabs( num ) > limit ||
                             ( exact && llabs( num ) >= factor ) ); ++i )
    { num /= factor; if( num % factor != 0 ) exact = false; p = prefix[i]; }
  snprintf( buf, sizeof buf, "%lld %s", num, p );
  return buf;
  }


int open_instream( const std::string & name, struct stat * const in_statsp,
                   const bool to_stdout, const bool reg_only )
  {
  int infd = open( name.c_str(), O_RDONLY | o_binary );
  if( infd < 0 )
    {
    if( verbosity >= 0 )
      std::fprintf( stderr, "%s: Can't open input file '%s': %s.\n",
                    program_name, name.c_str(), std::strerror( errno ) );
    }
  else
    {
    const int i = fstat( infd, in_statsp );
    const mode_t & mode = in_statsp->st_mode;
    const bool can_read = ( i == 0 && !reg_only &&
                            ( S_ISBLK( mode ) || S_ISCHR( mode ) ||
                              S_ISFIFO( mode ) || S_ISSOCK( mode ) ) );
    if( i != 0 || ( !S_ISREG( mode ) && ( !to_stdout || !can_read ) ) )
      {
      if( verbosity >= 0 )
        std::fprintf( stderr, "%s: Input file '%s' is not a regular file%s.\n",
                      program_name, name.c_str(),
                      ( can_read && !to_stdout ) ?
                      " and '--stdout' was not specified" : "" );
      close( infd );
      infd = -1;
      }
    }
  return infd;
  }


int open_outstream_rw( const std::string & output_filename,
                       const bool force )
  {
  int flags = O_CREAT | O_RDWR | o_binary;
  if( force ) flags |= O_TRUNC; else flags |= O_EXCL;

  int outfd = open( output_filename.c_str(), flags, all_rw );
  if( outfd < 0 && verbosity >= 0 )
    {
    if( errno == EEXIST )
      std::fprintf( stderr, "%s: Output file '%s' already exists."
                            " Use '--force' to overwrite it.\n",
                    program_name, output_filename.c_str() );
    else
      std::fprintf( stderr, "%s: Can't create output file '%s': %s.\n",
                    program_name, output_filename.c_str(), std::strerror( errno ) );
    }
  return outfd;
  }


void show_error( const char * const msg, const int errcode, const bool help )
  {
  if( verbosity >= 0 )
    {
    if( msg && msg[0] )
      {
      std::fprintf( stderr, "%s: %s", program_name, msg );
      if( errcode > 0 )
        std::fprintf( stderr, ": %s", std::strerror( errcode ) );
      std::fprintf( stderr, "\n" );
      }
    if( help && invocation_name && invocation_name[0] )
      std::fprintf( stderr, "Try '%s --help' for more information.\n",
                    invocation_name );
    }
  }


void internal_error( const char * const msg )
  {
  if( verbosity >= 0 )
    std::fprintf( stderr, "%s: internal error: %s.\n", program_name, msg );
  std::exit( 3 );
  }


int main( const int argc, const char * const argv[] )
  {
  int infd = -1;
  Mode program_mode = m_none;
  bool force = false;
  bool keep_input_files = false;
  bool to_stdout = false;
  std::string input_filename;
  std::string default_output_filename;
  std::string range_string;
  std::vector< std::string > filenames;
  invocation_name = argv[0];

  const Arg_parser::Option options[] =
    {
    { 'c', "stdout",           Arg_parser::no  },
    { 'd', "decompress",       Arg_parser::no  },
    { 'D', "range-decompress", Arg_parser::yes },
    { 'f', "force",            Arg_parser::no  },
    { 'h', "help",             Arg_parser::no  },
    { 'k', "keep",             Arg_parser::no  },
    { 'l', "list",             Arg_parser::no  },
    { 'm', "merge",            Arg_parser::no  },
    { 'o', "output",           Arg_parser::yes },
    { 'q', "quiet",            Arg_parser::no  },
    { 'R', "repair",           Arg_parser::no  },
    { 's', "split",            Arg_parser::no  },
    { 't', "test",             Arg_parser::no  },
    { 'v', "verbose",          Arg_parser::no  },
    { 'V', "version",          Arg_parser::no  },
    {  0 ,  0,                 Arg_parser::no  } };

  const Arg_parser parser( argc, argv, options );
  if( parser.error().size() )				// bad option
    { show_error( parser.error().c_str(), 0, true ); return 1; }

  int argind = 0;
  for( ; argind < parser.arguments(); ++argind )
    {
    const int code = parser.code( argind );
    if( !code ) break;					// no more options
    const std::string & arg = parser.argument( argind ).c_str();
    switch( code )
      {
      case 'c': to_stdout = true; break;
      case 'd': set_mode( program_mode, m_decompress ); break;
      case 'D': set_mode( program_mode, m_range );
                range_string = arg; break;
      case 'f': force = true; break;
      case 'h': show_help(); return 0;
      case 'k': keep_input_files = true; break;
      case 'l': set_mode( program_mode, m_list ); break;
      case 'm': set_mode( program_mode, m_merge ); break;
      case 'o': default_output_filename = arg; break;
      case 'q': verbosity = -1; break;
      case 'R': set_mode( program_mode, m_repair ); break;
      case 's': set_mode( program_mode, m_split ); break;
      case 't': set_mode( program_mode, m_test ); break;
      case 'v': if( verbosity < 4 ) ++verbosity; break;
      case 'V': show_version(); return 0;
      default : internal_error( "uncaught option" );
      }
    } // end process options

#if defined(__MSVCRT__) || defined(__OS2__)
  _fsetmode( stdin, "b" );
  _fsetmode( stdout, "b" );
#endif

  if( program_mode == m_none )
    {
    show_error( "You must specify the operation to be performed.", 0, true );
    return 1;
    }

  switch( program_mode )
    {
    case m_generate:
    case m_recover:
    case m_update:
    case m_none: internal_error( "invalid operation" ); break;
    case m_decompress: break;
    case m_list:
      one_file( argind, parser.arguments() );
      return list_file( parser.argument( argind ) );
    case m_merge:
      for( ; argind < parser.arguments(); ++argind )
        filenames.push_back( parser.argument( argind ) );
      if( filenames.size() < 2 )
        { show_error( "You must specify at least 2 files.", 0, true ); return 1; }
      if( !default_output_filename.size() )
        default_output_filename = insert_fixed( filenames[0] );
      return merge_files( filenames, default_output_filename, force );
    case m_range:
      one_file( argind, parser.arguments() );
      return range_decompress( parser.argument( argind ),
                               default_output_filename, range_string,
                               to_stdout, force );
    case m_repair:
      one_file( argind, parser.arguments() );
      if( !default_output_filename.size() )
        default_output_filename = insert_fixed( parser.argument( argind ) );
      return repair_file( parser.argument( argind ), default_output_filename, force );
    case m_split:
      one_file( argind, parser.arguments() );
      return split_file( parser.argument( argind ), default_output_filename, force );
    case m_test: break;
    }

  if( program_mode == m_test )
    outfd = -1;
  else if( program_mode != m_decompress )
    internal_error( "invalid decompressor operation" );

  bool filenames_given = false;
  for( ; argind < parser.arguments(); ++argind )
    {
    if( parser.argument( argind ) != "-" ) filenames_given = true;
    filenames.push_back( parser.argument( argind ) );
    }

  if( filenames.empty() ) filenames.push_back("-");
  if( !to_stdout && program_mode != m_test &&
      ( filenames_given || default_output_filename.size() ) )
    set_signals();

  Pretty_print pp( filenames, verbosity );

  int retval = 0;
  for( unsigned int i = 0; i < filenames.size(); ++i )
    {
    struct stat in_stats;
    output_filename.clear();

    if( !filenames[i].size() || filenames[i] == "-" )
      {
      input_filename.clear();
      infd = STDIN_FILENO;
      if( program_mode != m_test )
        {
        if( to_stdout || !default_output_filename.size() )
          outfd = STDOUT_FILENO;
        else
          {
          output_filename = default_output_filename;
          outfd_mode = all_rw;
          if( !open_outstream( force ) )
            {
            if( outfd == -1 && retval < 1 ) retval = 1;
            close( infd ); infd = -1;
            continue;
            }
          }
        }
      }
    else
      {
      input_filename = filenames[i];
      const int eindex = extension_index( input_filename );
      infd = open_instream( input_filename, &in_stats, to_stdout );
      if( infd < 0 ) { if( retval < 1 ) retval = 1; continue; }
      if( program_mode != m_test )
        {
        if( to_stdout ) outfd = STDOUT_FILENO;
        else
          {
          set_d_outname( input_filename, eindex );
          outfd_mode = usr_rw;
          if( !open_outstream( force ) )
            {
            if( outfd == -1 && retval < 1 ) retval = 1;
            close( infd ); infd = -1;
            continue;
            }
          }
        }
      }

    if( !check_tty( infd ) ) return 1;

    if( output_filename.size() && !to_stdout && program_mode != m_test )
      delete_output_on_interrupt = true;
    const struct stat * const in_statsp = input_filename.size() ? &in_stats : 0;
    pp.set_name( input_filename );
    const int tmp = decompress( infd, pp, program_mode == m_test );
    if( tmp > retval ) retval = tmp;
    if( tmp && program_mode != m_test ) cleanup_and_fail( retval );

    if( delete_output_on_interrupt )
      close_and_set_permissions( in_statsp );
    if( input_filename.size() )
      {
      close( infd ); infd = -1;
      if( !keep_input_files && !to_stdout && program_mode != m_test )
        std::remove( input_filename.c_str() );
      }
    }
  if( outfd >= 0 && close( outfd ) != 0 )
    {
    show_error( "Can't close stdout", errno );
    if( retval < 1 ) retval = 1;
    }
  return retval;
  }
